/*{{{  File Header*/
/*
 *     Program:       itconfig
 *     Module:        %M%    %W% %G% 
 *
 *
 *     Copyright Inmos Limited 1990.
 *
 *     Purpose:       routines for reading and writing data to remote
 *                    TCP/IP system and loading configuration data from various
 *                    sources for remote configuration program for TCP/IP
 *                    transputer LAN interface products.
 *
 *     Compatability: BSD UNIX (host) and Inmos 1.0 Beta UDP/IP (target node)
 *                    VAX/VMS running WIN/TCP
 *                    PC NFS
 *
 *     History
 *        GJT     01/05/90      a start!
 *        GJT     16/05/90      updates for latest spec
 *        GJT     31/10/90      updates for alpha release
 *        GJT     15/02/91      Changed all occurences of EEPROM to NVRAM
 *        GJT     18/03/91      Changes for new editing mechanism
 *        GJT     05/04/91      Fixed edit type TOS bug and net file name 
 *                              truncate bug. Added default names
 *        DAE     28/06/91      Added DEC/TCP support
 *        NMGS    17/01/92      Added PC-TCP support, renamed get_addr
 *                              to parse_addr to prevent symbol clash
 *        NMGS    22/03/92      Upgraded PC-NFS toolkit support to v2.0
 */

/*}}}  */

/*{{{  include files*/
#include <stdio.h>
#include <string.h>
#include <netdb.h>

#ifdef DECTCP
#include <types.h>
#include <socket.h>
#include <in.h>
#include <netdb.h>
#else
#include <assert.h>
#ifdef PCNFS
#include <sys/tk_types.h>
#include <netinet/in.h>
#include <in_addr.h>
#else
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif /* PCNFS */
#include <sys/socket.h>
#ifndef WINTCP
#include <memory.h>
#endif /* WINTCP */
#endif /* DECTCP */

#include "itconfig.h" 
#include "types.h" /**/
#include "debug.h" /* Linkops config manager library include file */
#include "opserror.h" /* Linkops config manager library include file */
#include "lmethod.h" /* Linkops config manager library include file */

/*}}}  */


/*{{{  local typedefs*/

/*
 * Typedefs for this file
 */

/*
 * Frame forconfiguration protocol
 */

typedef struct {
                unsigned char vstring[VFIELD_SIZE];
                unsigned char ethernet_address[ETHER_ADDR_LEN];
                unsigned char id;
                unsigned char op;
                short int reserved;
                short int data_len;
                unsigned char data[MAX_DATA_LEN];
               } frame;


/*}}}  */

/*{{{  global variables*/
/*
 * Global variables  (external)
 */

extern unsigned int ether_addr[6]; /* Ethernet address of transputer TCP/IP system */

extern char id;  /* Id for comms with transputer TCP/ip system */

extern BOOL edited_config;  /* Set to TRUE if parameters are changed */

extern int SPEED_TABLE[NO_SPEEDS]; /* Holds link speeds */


extern int skt;        /* socket id for comms with transputer TCP/IP system */
extern struct sockaddr_in skt_in, rsin;
extern struct netent *net;

/*}}}  */

/*{{{  external function defs*/
#include "conlib.h"
/*
 * External function defs
 */

extern void           debug_m();
extern void           tidy_string ();
extern void           type_help ();

#ifdef VMS
extern void           usleep ();
#endif
extern int            timed_read ();

extern char * getenv ();
/*}}}  */

/*{{{  user i/o routines for string handling etc.*/
/*{{{  tidy string*/
void tidy_string (string)
unsigned char *string;
/*
 * Truncate a string which must have a '\0' at end or a space or a newline
 */
{
  for (; (*string != ' ') && (*string != '\n') && (*string != '\0'); string++)
    ;
  *string = '\0';

  return;
}
/*}}}  */
/*{{{  tidy_network*/
void tidy_network (config)
unsigned char config[];
{
  int i, index;
  BOOL rubbish;

  if (((int) config[SYSTEM_NAME_OFFSET]) > MAX_NAME_LEN)
  {
    printf("Warning: corrupt system name\n");
    config[SYSTEM_NAME_OFFSET] = (unsigned char) MAX_NAME_LEN;
  }

  /* 
   * As domain name can be up to 255 chars it is always ok
   */
  
  /*
   * Check first router entry if it is all 0's or 255's assume it is 
     corrupt and truncate it at zero entries.
   */
  index = ROUTER_OFFSET;
  rubbish = TRUE;
  for (i = 0; i < ROUTER_ENTRY_SIZE; i++)
  {
    if ((config[index] != (unsigned char) 0X00) &&  (config[index] != (unsigned char) 0XFF))
    {
      rubbish = FALSE;  
    }
  }

  if (rubbish == TRUE)
  {
    printf("Warning: corrupt router table, all entries removed\n");
    config[ROUTER_OFFSET] = (unsigned char) END_OF_ROUTER_TABLE;
  }

  return;
}
/*}}}  */
/*{{{  tidy_resource*/
void tidy_resource (config)
unsigned char config[];
{
  int index, i;
  BOOL entry_truncated = FALSE;

  index = RESOURCE_LEN_O;
  for (i = 0; i < MAX_RESOURCE_ENTRIES; i++)
  {
    if (config[index] > (unsigned char) MAX_NAME_LEN)
    {
      config[index] = (unsigned char) MAX_NAME_LEN;
      entry_truncated = TRUE;
   
    }

    index += RESOURCE_ENTRY_SIZE;
  
  }

  if (entry_truncated == TRUE)
  {
    printf("Warning: corrupt resource data\n");
  }

  return;
}
/*}}}  */

/*{{{  hex_colon_addr*/
unsigned char *hex_colon_addr (num, bytes)
int num;
unsigned int bytes[];
/*
 * Return Ethenet address as hex, colon sepated string
 */
{
    static unsigned char op[MAX_DOTBYTES * 3 + 1];

    if (num > MAX_DOTBYTES)
        return (NULL);

     sprintf((char *) op, "%02x:%02x:%02x:%02x:%02x:%02x\0", bytes[0] & 0XFF,
           bytes[1] & 0xFF, bytes[2] & 0XFF, bytes[3] & 0XFF,
            bytes[4] & 0XFF, bytes[5] & 0XFF);

    return (op);
}
/*}}}  */
/*{{{  dec_dot_addr*/
unsigned char *dec_dot_addr (num, bytes)
int num;
unsigned int bytes[];
/*
 * Return IP address as hex, dot sepated string
 */

{
    static unsigned char op[MAX_COLONBYTES * 4 + 1];

    if (num > MAX_COLONBYTES)
        return (NULL);

    sprintf((char *) op, "%u.%u.%u.%u\0", bytes[0], bytes[1], bytes[2], bytes[3]);

    return (op);
}

/*}}}  */
/*{{{  ip_addr_copy*/
void ip_addr_copy (string, numbers)
unsigned char *string;
unsigned int numbers[IP_ADDR_LEN];
/*
 * Copy address unsigned int numbers to byte string
 */
{
  int i;

  for (i = 0; i < IP_ADDR_LEN; i++, string++)
     *string = (char) (numbers[i] & 0XFF);

  return;
}
/*}}}  */
/*{{{  addr_toword*/
void addr_toword (numbers, string, len)
unsigned int *numbers;
unsigned char *string;
int len;
/*
 * Copy address numbers in string to unsigned int array
 */
{
  int i;

  for (i = 0; i < len; i++, string++, numbers++)
     *numbers = ((unsigned int) *string) & 0XFF;

  return;
}
/*}}}  */

/*{{{  index_to_speed*/
int index_to_speed (index)
unsigned char index;
{
  if ((index < 0) || (index >= sizeof(SPEED_TABLE)))
    return(0);
  else
    return(SPEED_TABLE[(int) index]);

}
/*}}}  */
/*{{{  speed_to_index*/
int speed_to_index (speed)
int speed;
{
  int index = 0;


  while ((SPEED_TABLE[index] != speed) && (index < sizeof(SPEED_TABLE)))
   index++;

  if (index == sizeof(SPEED_TABLE))
    index = (-1);

  return(index);
}
/*}}}  */

/*{{{  show_type*/
char *show_type (type)
unsigned int type;
{
  static char type_string[MAX_TYPE_SLEN];
  int index;

  if ((type & DEFAULT_TYPE) == DEFAULT_TYPE)
  {
    strcpy(type_string, DEFAULT_TYPE_S);
    index = sizeof(DEFAULT_TYPE_S);
  }
  else if ((type & NETWORK_TYPE) == NETWORK_TYPE)  
  { 
    strcpy(type_string, NETWORK_TYPE_S);
    index = sizeof(NETWORK_TYPE_S);
  }
  else  
  {
    strcpy(type_string, HOST_TYPE_S);
    index = sizeof(HOST_TYPE_S);
  }

  /* if ((type & DEFAULT_TYPE) == NETWORK_TYPE)
    strcpy(&type_string[index], NO_TOS_S);
  else
    strcpy(&type_string[index], TOS_S); */

  return(type_string);
}
/*}}}  */
/*{{{  show_entry*/
void show_entry (number, entry)
int number;
struct router_entry *entry;
{

  printf("\n%d: %s, ", number, show_type(entry->type));
  if ((entry->type & DEFAULT_TYPE) != DEFAULT_TYPE)
    printf("%s, ", dec_dot_addr(IP_ADDR_LEN, entry->address));

  if ((entry->type & NO_TOS_TYPE) != NO_TOS_TYPE)
    printf("%u, ", entry->TOS);
  printf("%s\n", dec_dot_addr(IP_ADDR_LEN, entry->gateway));

  return;
}
/*}}}  */

/*{{{  get_router_type*/
BOOL get_router_type (type, show_default)
unsigned int *type;
BOOL show_default;
{
  char reply[COMMAND_LEN];
  int new_type;
  int new_tos;
  BOOL ok = TRUE;
  BOOL changed = FALSE;
  BOOL reading;

  reading = TRUE;
  /*{{{  get type*/
  while ((reading == TRUE) && (ok == TRUE))
  {
    if (show_default == TRUE)
      printf("Type (%s) [D(efault), H(ost), N(etwork)]: ",
              show_type(*type));
    else
      printf("Type (N) [D(efault), H(ost), N(etwork), Q(uit), ?]: ");
    fgets(reply, sizeof(reply), stdin);
    tidy_string(reply);
    switch (reply[0])
    {
      case '\0':    /* Default selected */
        if (show_default == FALSE)
          new_type = NETWORK_TYPE;
        else
          new_type = (*type) & TOS_MASK; /* Mask out TOS not required bit */
        reading = FALSE;
        break;
      case 'Q': case 'q':
        ok = FALSE;
        break;
      case '?':
        type_help();
        break;
      case 'd': case 'D':
        changed = TRUE;
        new_type = DEFAULT_TYPE;
        reading = FALSE;
        break;
      case 'h': case 'H':
        changed = TRUE;
        new_type = HOST_TYPE;
        reading = FALSE;
        break;
      case 'n': case 'N':
        changed = TRUE;
        new_type = NETWORK_TYPE;
        reading = FALSE;
        break;
      default:
        break;  /* SKIP */
    }
  }
  /*}}}  */

  reading = TRUE;
  /*{{{  get TOS required*/
  while ((reading == TRUE) && (ok == TRUE))
  {
    if (show_default == TRUE)
      if ((*type & NO_TOS_TYPE) == NO_TOS_TYPE)
        printf("Specify TOS (N) [N(o), Y(es)]: ");
      else
        printf("Specify TOS (Y) [N(o), Y(es)]: ");
    else
      printf("Specify TOS (N) [N(o), Y(es)]: ");
    fgets(reply, sizeof(reply), stdin);
    tidy_string(reply);
    switch (reply[0])
    {
      case '\0':    /* Default selected */
        if (show_default == FALSE)
        {
          new_tos = NO_TOS_TYPE;
        }
        else
        {
          new_tos = *type & NO_TOS_TYPE;  /* Set to old TOS REQ'd value */
        }
        reading = FALSE;
        break;
      case 'y':
      case 'Y':
        changed = TRUE;
        new_tos = 0;
        reading = FALSE;
        break;
      case 'n':
      case 'N':
        changed = TRUE;
        new_tos = NO_TOS_TYPE;
        reading = FALSE;
        break;
      default:
        break;  /* SKIP */
    }
  }
  /*}}}  */

  if ((show_default == TRUE) && (changed == TRUE) && (ok == TRUE))
    *type = new_type | new_tos | VALID_ENTRY ;
  else if (show_default == TRUE)
  {
    ok = FALSE;
  }
  else if (ok == TRUE)
    *type = new_type | new_tos | VALID_ENTRY;

  return(ok);
}
/*}}}  */
/*{{{  parse_addr*/
BOOL parse_addr (addr, name, show_default)
int addr[];
char *name;
BOOL show_default;
{
  BOOL ok;
  int len;
  char addr_line[MAX_NAME_LEN + 1];
  char dummy;

  do
  {
    if (show_default == TRUE)
      printf("%s (%s): ", name, dec_dot_addr(IP_ADDR_LEN, addr));
    else
      printf("%s: ", name);
    fgets(addr_line, sizeof(addr_line), stdin);
    tidy_string(addr_line);
#ifdef VMS
    len = sscanf( addr_line, "%d.%d.%d.%d%c", &addr[0], &addr[1], &addr[2], &addr[3], &dummy);
#else
    len = sscanf( addr_line, "%u.%u.%u.%u%c", &addr[0], &addr[1], &addr[2], &addr[3], &dummy);
#endif

    if ((len != IP_ADDR_LEN) && (addr_line[0] != '\0'))
    {
      struct hostent *hp;

      hp = gethostbyname(addr_line);
      if (hp != (struct hostent *) NULL)
      {
        addr_toword(addr, hp->h_addr, IP_ADDR_LEN);
        len = IP_ADDR_LEN;
        printf("IP address of %s is %s\n", addr_line,  dec_dot_addr(IP_ADDR_LEN, addr));
      }
      else
      {
        printf("Error: host %s not known\n", addr_line);
      }

    }

  } while ((addr_line[0] != '\0') && (len != IP_ADDR_LEN));

  if (len != IP_ADDR_LEN)
    ok = FALSE;
  else
    ok = TRUE;

  return(ok);
}
/*}}}  */
/*{{{  get_config_addr*/
void get_config_addr (config_ptr, name)
unsigned char *config_ptr;
char *name;
{
  unsigned int addr[IP_ADDR_LEN];
  BOOL ok;

  addr_toword(addr, config_ptr, IP_ADDR_LEN);
  ok = parse_addr(addr, name, TRUE);
  if (ok == TRUE)
  {
    edited_config = TRUE;
    ip_addr_copy(config_ptr, addr);
  }

  return;
}
/*}}}  */
/*{{{  get_config_name*/
void get_config_name (config_ptr, opt_name, max_length, def_name)
unsigned char *config_ptr;
char *opt_name;
int max_length;
unsigned char *def_name;
{
  int i;
  char name[MAX_DOMAIN_LEN + 1]; /* The largest name that will be asked for */
  unsigned int addr[IP_ADDR_LEN];
  char addr_line[MAX_DOMAIN_LEN + 1];
  int len;

  len = (int) *config_ptr & 0XFF;
  if (len == 0)
  { /* No name present, use supplied default */
    strcpy(name, def_name);
  }
  else
  {
    memcpy(name, (char *) (config_ptr + 1), len);
    name[len] = '\0';
  }
  printf("%s (%s): ", opt_name, name);
  fgets(name, max_length, stdin);
  tidy_string(name);
  if ((name[0] == '\0') && (len == 0)) 
  /* 
   * Accept default and no name - copy in supplied default.
   */
  {
    strcpy(name, def_name);
  }

  if (name[0] != '\0') /* New name */
  {
    for (i = 0; name[i] != '\0'; i++);  /* find length */
    *config_ptr = (unsigned char) i;
    memcpy((char *) (config_ptr + 1), name, i);
    for (; i < max_length; i++)
      *(config_ptr + (i + 1)) = '\0';
    edited_config = TRUE;
  }

  return;
}
/*}}}  */
/*{{{  get_int*/
BOOL get_int (value, name, show_default)
unsigned int *value;
char *name;
BOOL show_default;
{
  BOOL ok;
  int len;
  char addr_line[MAX_NAME_LEN + 1];
  unsigned int new_value;
  char dummy;

  do
  {
    if (show_default == TRUE)
      printf("%s (%u): ", name, *value);
    else
      printf("%s: ", name);
    fgets(addr_line, sizeof(addr_line), stdin);
    tidy_string(addr_line);
#ifdef VMS  /* VMS sscanf bug */
    len = sscanf( addr_line, "%d%c", &new_value, &dummy);
#else
    len = sscanf( addr_line, "%u%c", &new_value, &dummy);
#endif
  } while ((addr_line[0] != '\0') && (len != 1));

  if (len != 1)
    ok = FALSE;
  else
  {
    ok = TRUE;
    *value = new_value;
  }

  return(ok);
}
/*}}}  */
/*}}}  */

/*{{{  init_data_block*/
void init_data_block (data_block, size)
  unsigned char *data_block;
  int size;
/*
 *  Initialise a char data block of given size to nulls
 */
{
  int i;

  for (i = 0; i < size; i++, data_block++)
    *data_block = '\0';

  return;
}
/*}}}  */

/*{{{  NVRAM write*/

BOOL NVRAM_write (ptr, address, length)
unsigned char *ptr;
unsigned short int address;
short int length;
/*
 * Write data to NVRAM
 */
{
  int i;
  BOOL got_frame;
  frame req_frame, reply_frame;
  short int *req_addr = (short int *) req_frame.data;
  unsigned char *d_ptr = &req_frame.data[2]; /* point to after address field */

  /*
   * Set up outgoing frame
   */

  strcpy(req_frame.vstring, verify_string);
  for (i = 0; i < ETHER_ADDR_LEN; i++)
    req_frame.ethernet_address[i] = (unsigned char) (ether_addr[i] & 0XFF);
  req_frame.id = id;
  req_frame.op = WRITE_REQ;
  req_frame.reserved = 0;
  req_frame.data_len = htons(length + 2);
  *req_addr = htons(address);
  memcpy(d_ptr, ptr, length);
   

  /*
   * initialise values for returned frame
   */

  reply_frame.op = NOP;
  reply_frame.id = ! id;


  i = 0;
  got_frame = FALSE;
  while ((i < MAX_RETRIES) && (got_frame == FALSE))
  {
    int read_len;

    if (sendto(skt, (char *) &req_frame, (length + 2) + MIN_FRAME_SIZE, 0,
               (struct sockaddr *) &rsin, sizeof (struct sockaddr)) < 0)
    {
      perror ("Error: NVRAM write");
      exit(ERROR_EXIT);
    }

    read_len = timed_read(skt, (char *) &reply_frame, MAX_DATAGRAM_LEN, 500L); 
    if (read_len < 1)
    {
      reply_frame.op = htons(NOP);
      reply_frame.id = ! id;
    }
    else  if (read_len >= MIN_FRAME_SIZE)   /* Check the frame */
    {
      int compare;

      compare = memcmp(verify_string, reply_frame.vstring, strlen(verify_string));
      if ((compare == 0) && (reply_frame.op == WRITE_REPLY) && 
          (reply_frame.id == id))
        got_frame = TRUE;

    }

    i++;

  }

  id++;

  return(got_frame);
}

/*}}}  */

/*{{{  NVRAM read*/
int NVRAM_read (ptr, address, length)
unsigned char *ptr;
unsigned short int address;
short int length;
/*
 * Read data from NVRAM
 */
{
  int i;
  BOOL got_frame;
  short int req_data[2];
  int bytes_read = 0;
  frame req_frame, reply_frame;

  /*
   * Set up outgoing frame
   */

  strcpy(req_frame.vstring, verify_string);
  for (i = 0; i < ETHER_ADDR_LEN; i++)
    req_frame.ethernet_address[i] = (char) (ether_addr[i] & 0XFF);
  req_frame.id = id;
  req_frame.op = EREAD_REQ;
  req_frame.reserved = 0;
  req_frame.data_len = htons(4);
  req_data[0] = htons(address);
  req_data[1] = htons(length);
  memcpy(req_frame.data, (char *)req_data, 4);

  /*
   * initialise values for returned frame
   */

  reply_frame.op = NOP;
  reply_frame.id = ! id;


  i = 0;
  got_frame = FALSE;
  while ((i < MAX_RETRIES) && (got_frame == FALSE))
  {
    int read_len;

    if (sendto(skt, (char *) &req_frame, MIN_FRAME_SIZE + 4, 0,
               (struct sockaddr *) &rsin, sizeof (struct sockaddr)) < 0)
    {
      perror ("Error: NVRAM read");
      exit(ERROR_EXIT);
    }

    read_len = timed_read(skt, (char *) &reply_frame, MAX_DATAGRAM_LEN, 500L); 
    if (read_len < 1)
    {
      reply_frame.op = htons(NOP);
      reply_frame.id = ! id;
    }
    else  if (read_len >= MIN_FRAME_SIZE)   /* Check the frame */
    {
      int compare;

      compare = memcmp(verify_string, reply_frame.vstring, strlen(verify_string));
      if ((compare == 0) &&  (reply_frame.op == EREAD_REPLY) && 
          (reply_frame.id == id))
        got_frame = TRUE;
      
    }


    i++;

  }

  if (got_frame)
  {
    bytes_read = htons(reply_frame.data_len);
    memcpy(ptr, reply_frame.data, bytes_read);
  }

  id++;

  return(bytes_read);
}

/*}}}  */

/*{{{  RAM read*/
int RAM_read (ptr, address, length)
unsigned char *ptr;
unsigned short int address;
short int length;
/*
 * Read data from RAM
 */
{
  int i;
  BOOL got_frame;
  short int req_data[2];
  int bytes_read = 0;
  frame req_frame, reply_frame;

  /*
   * Set up outgoing frame
   */

  strcpy(req_frame.vstring, verify_string);
  for (i = 0; i < ETHER_ADDR_LEN; i++)
    req_frame.ethernet_address[i] = (char) (ether_addr[i] & 0XFF);
  req_frame.id = id;
  req_frame.op = MREAD_REQ;
  req_frame.reserved = 0;
  req_frame.data_len = htons(4);
  req_data[0] = htons(address);
  req_data[1] = htons(length);
  memcpy(req_frame.data, (char *)req_data, 4);

  /*
   * initialise values for returned frame
   */

  reply_frame.op = NOP;
  reply_frame.id = ! id;


  i = 0;
  got_frame = FALSE;
  while ((i < MAX_RETRIES) && (got_frame == FALSE))
  {
    int read_len;

    if (sendto(skt, (char *) &req_frame, MIN_FRAME_SIZE + 4, 0,
               (struct sockaddr *) &rsin, sizeof (struct sockaddr)) < 0)
    {
      perror ("Error: RAM read");
      exit(ERROR_EXIT); 
    }

    read_len = timed_read(skt, (char *) &reply_frame, MAX_DATAGRAM_LEN, 500L); 
    if (read_len < 1)
    {
      reply_frame.op = htons(NOP);
      reply_frame.id = ! id;
    }
    else  if (read_len >= MIN_FRAME_SIZE)   /* Check the frame */
    {
      int compare;

      compare = memcmp(verify_string, reply_frame.vstring, strlen(verify_string));
      if ((compare == 0) &&
          (reply_frame.op == MREAD_REPLY) && (reply_frame.id == id))
      { 
        got_frame = TRUE;
      }

    }

    i++;

  }

  if (got_frame)
  {
    bytes_read = htons(reply_frame.data_len);
    memcpy(ptr, reply_frame.data, bytes_read);
  }

  id++;

  return(bytes_read);
}


/*}}}  */

/*{{{  read_net_NVRAM*/
int read_net_NVRAM (config)
unsigned char config[];

/*
 * Simple routine to read all network config data from NVRAM
 */
{
  unsigned short int address = 0;
  int bytes_read = 0;
  
  init_data_block(config, NET_DATA_SIZE);
  bytes_read =  NVRAM_read(&config[0], address, DATA_LEN);
  address += (unsigned short int) bytes_read;

  while (bytes_read == DATA_LEN)
  {
     int bytes_to_read;

     if ((NET_DATA_SIZE - ((int) address)) > DATA_LEN)
     {
       bytes_to_read = DATA_LEN;
     }
     else
     {
       bytes_to_read = NET_DATA_SIZE - (int) address;
     }
     bytes_read =  NVRAM_read(&config[address], address, bytes_to_read);
     address += (unsigned short int) bytes_read;
  }

  if (address < NET_DATA_SIZE)
  {  
   /* If router table not all available it is ignored */
    config[ROUTER_OFFSET] = END_OF_ROUTER_TABLE;
  }

  tidy_network(config);

  return((int) address);
}

/*}}}  */

/*{{{  read net RAM*/
int read_net_RAM (config)
unsigned char config[];

/*
 * Read network part of configuration from RAM
 */
{
  unsigned short int address = 0;
  int bytes_read = 0;

  init_data_block(config, NET_DATA_SIZE);
  bytes_read =  RAM_read(&config[0], address, DATA_LEN);
  address += (unsigned short int) bytes_read;

  while (bytes_read == DATA_LEN)
  {
     int bytes_to_read;

     if ((NET_DATA_SIZE - ((int) address)) > DATA_LEN)
     {
       bytes_to_read = DATA_LEN;
     }
     else
     {
       bytes_to_read = NET_DATA_SIZE - (int) address;
     }
     bytes_read = RAM_read(&config[address], address, bytes_to_read);
     address += (unsigned short int) bytes_read;
  }

  if (address < NET_DATA_SIZE)
    config[ROUTER_OFFSET] = END_OF_ROUTER_TABLE;
  /* If router table not all available it is ignored */

  tidy_network(config);

  return((int) address);

}


/*}}}  */

/*{{{  read net file*/
int read_net_file (config, filename)
unsigned char config[];
char *filename;

/*
 * Read network configuration data from named file
 */
{
  FILE *fid;
  int index = 0;
  unsigned char *ptr;
  BOOL error = FALSE;

  fid = fopen(filename, "r");
  if (fid == NULL)
  {
    error = TRUE;
    printf("Error: unable to open network data file %s for reading\n", filename);
  }
  else
  {
    char line[MAX_NET_LINE_SIZE];
    unsigned int addr[IP_ADDR_LEN];

    fgets(line, MAX_NET_LINE_SIZE, fid);
    if (strncmp(netfile_version, line, NETFILE_LIMIT) == 0)
    {
      unsigned int lops_port;
      unsigned int link_speed;
      int i;
      int count;
      int line_count = 1;
      int router_entries = 0;
      int reading_router = TRUE;
      init_data_block(config, NET_DATA_SIZE);
      fgets(line, sizeof(line), fid); /* Swallow comment line */
      line_count++;

      for (i = 0; i < ETHER_ADDR_LEN; i++, index++)
        config[index] = ether_addr[i] & 0XFF;
      index += 2; /* Skip over padding */
      /*
       * Loop for addr, subnet mask, broadcast, primary & secondary server
       */
      do
      {
        ptr = (unsigned char *) fgets(line, sizeof(line), fid);
        if (ptr != NULL)
        {
#ifdef VMS  /* VMS sscanf bug */
          count = sscanf (line, "%d.%d.%d.%d%", &addr[0], &addr[1],
                                        &addr[2], &addr[3]);
#else
          count = sscanf (line, "%u.%u.%u.%u", &addr[0], &addr[1],
                                        &addr[2], &addr[3]);
#endif
          ip_addr_copy(&config[index], addr);
          line_count++;
          index += IP_ADDR_LEN;
        }
      } while ((ptr != NULL) &&(line_count < (NAME_LINE_NO - 2)) && (count == IP_ADDR_LEN));

      error = (count != IP_ADDR_LEN) | (ptr == NULL);

      /* Get linkops port */
      if (error == FALSE)
      {
        ptr = (unsigned char *) fgets(line, sizeof(line), fid);
        if (ptr != NULL)
        {
#ifdef VMS  /* VMS sscanf bug */
          count = sscanf (line, "%d", &lops_port);
#else
          count = sscanf (line, "%u", &lops_port);
#endif
          if (count == 1)
          {
            config[index++] = (unsigned char) ((lops_port & 0XFF00) >> 8);  /* MS byte */
            config[index++] = (unsigned char) (lops_port & 0XFF);  /* LS byte */
            line_count++;
          }

          else
          {
            error = TRUE;
          }
        }
        else
        {
          error = TRUE;
        }
      }

      /* Get link speed */
      if (error == FALSE)
      {
        ptr = (unsigned char *) fgets(line, sizeof(line), fid);
        if (ptr != NULL)
        {
#ifdef VMS  /* VMS sscanf bug */
          count = sscanf (line, "%d", &link_speed);
#else
          count = sscanf (line, "%u", &link_speed);
#endif
          if (count == 1)
          {
            config[index++] = (unsigned char) (speed_to_index(link_speed) & 0XFF);
            config[index] = (unsigned char) 0;
            line_count++;
          }

          else
          {
            error = TRUE;
          }
        }
        else
        {
          error = TRUE;
        }
      }

      /*{{{  host and domain name*/
      if (error == FALSE)
      {
        int len;
        /* Host name */
      
        index = SYSTEM_NAME_OFFSET;
        ptr = (unsigned char *) fgets(line, sizeof(line), fid);
        if (ptr != NULL)
        /*{{{  read ok*/
        {
          tidy_string(line);
          for (len = 0; line[len] != '\0'; len++)
            ; /* Calculate length of name */
          if (len > MAX_NAME_LEN)
          {
            len = MAX_NAME_LEN;
            line[len] = '\0';
          }
          config[index] = (unsigned char) len;
          index++;
          strcpy(&config[index], line);
          line_count++;
        
          /* Domain name */
          index = DOMAIN_NAME_OFFSET;
          ptr = (unsigned char *) fgets(line, sizeof(line), fid);
          if (ptr != NULL)
          /*{{{  read ok*/
          {
            tidy_string(line);
            for (len = 0; line[len] != '\0'; len++)
              ; /* Calculate length of name */
            if (len > MAX_DOMAIN_LEN)
            {
              len = MAX_DOMAIN_LEN;
              line[len] = '\0';
            }
            config[index] = (unsigned char) len;
            index++;
            strcpy(&config[index], line);
            line_count++;
          }
          /*}}}  */
          /*{{{  otherwise, error*/
          else
          {
            error = TRUE;
          }
          /*}}}  */
        }
        /*}}}  */
        /*{{{  otherwise error*/
        else
        {
          error = TRUE;
        }
        /*}}}  */
      }
      /*}}}  */

      if (error == FALSE)
      {
        /*
         * Move to start of router table
         */

        index = ROUTER_OFFSET;

        while (reading_router) /* get router data */
        {
          unsigned int gate[IP_ADDR_LEN];
          unsigned int type;
          unsigned int tos;

          ptr = (unsigned char *) fgets(line, sizeof(line), fid);
          line_count++;
          if (ptr != NULL)
          {
#ifdef VMS  /* VMS sscanf bug */
            count = sscanf(line, "%d %d.%d.%d.%d %d %d.%d.%d.%d",
                    &type, &addr[0], &addr[1], &addr[2], &addr[3], &tos,
                    &gate[0], &gate[1], &gate[2], &gate[3]);
#else
            count = sscanf(line, "%u %u.%u.%u.%u %u %u.%u.%u.%u",
                    &type, &addr[0], &addr[1], &addr[2], &addr[3], &tos,
                    &gate[0], &gate[1], &gate[2], &gate[3]);
#endif
          }

          if (ptr == NULL)
          /*{{{  end of file*/
          {
            reading_router = FALSE;
            config[index] = END_OF_ROUTER_TABLE;
            index += ROUTER_ENTRY_SIZE;
            router_entries++;
          }
          /*}}}  */
          else if (count != 10)
          /*{{{  not in correct format*/
          {
            reading_router = FALSE;
            printf("Error: %s (%d) is a bad format network data file\n",
                    filename, line_count);
            config[index] = END_OF_ROUTER_TABLE;
            index += ROUTER_ENTRY_SIZE;
            router_entries++;
            error = TRUE;
          }
          /*}}}  */
          else if (type == END_OF_ROUTER_TABLE)
          /*{{{  end of router table*/
          {
            reading_router = FALSE;
            config[index] = END_OF_ROUTER_TABLE;
            index += ROUTER_ENTRY_SIZE;
            router_entries++;
          }
          /*}}}  */
          else if (router_entries >= MAX_ROUTER_ENTRIES)
          /*{{{  too many entries*/
          {
            reading_router = FALSE;
            printf("Warning: file %s contains too many router table entries\n",
                    filename);
            config[index] = END_OF_ROUTER_TABLE;
            index += ROUTER_ENTRY_SIZE;
            router_entries++;
          }
          /*}}}  */
          else
          /*{{{  otherwise, process entry*/
          {
            router_entries++;
            config[index] = (unsigned char) (type & TYPE_MASK);
            ip_addr_copy(&config[index + 1], addr);
            config[index + 5] = (unsigned char) (tos & 0XFF);
            config[index + 6] = (unsigned char) ((tos >> 8) & 0XFF);
            ip_addr_copy(&config[index + 7], gate);
            index += ROUTER_ENTRY_SIZE;
          }
          /*}}}  */

        }

      }
      else        /* ERROR */
      {
        printf("Error: %s (%d) is a bad format network data file\n",
                filename, line_count);
        error = TRUE;
        config[ROUTER_OFFSET] = END_OF_ROUTER_TABLE;
      }

    }
    else
    {
      printf("Error: %s is not a valid network data file\n", filename);
      config[ROUTER_OFFSET] = END_OF_ROUTER_TABLE;
      error = TRUE;
   }
   if (fclose(fid) != 0)
    printf("Error: closing network data file %s\n", filename);

  }
  if (error)
  {
    index = 0;
  }
  return(index);

}
/*}}}  */

/*{{{  read resource NVRAM*/

int read_resource_NVRAM (config)
unsigned char config[];

/*
 * Simple routine to read all resource config data from NVRAM
 */
{
  unsigned short int address = NET_DATA_SIZE;
  int bytes_read = 0;
  int length = 0;

  init_data_block(config, RESOURCE_DATA_BLOCK);

  bytes_read =  NVRAM_read(&config[length], address, DATA_LEN);

  while (bytes_read == DATA_LEN)
  {
    length += bytes_read;
    address += (unsigned short int) bytes_read;
    bytes_read =  NVRAM_read(&config[length], address, DATA_LEN);
  }
  length += bytes_read;

  tidy_resource(config);

  return(length);

}

/*}}}  */

/*{{{  read resource RAM*/
int read_resource_RAM (config)
unsigned char config[];

/*
 * Read resource part of configuration from RAM
 */
{
  unsigned short int address = NET_DATA_SIZE;
  int bytes_read = 0;
  int length = 0;

  init_data_block(config, RESOURCE_DATA_BLOCK);
  bytes_read =  RAM_read(&config[length], address, DATA_LEN);

  while (bytes_read == DATA_LEN)
  {
    length += bytes_read;
    address += (unsigned short int) bytes_read;
    bytes_read =  RAM_read(&config[length], address, DATA_LEN);
  }
  length += bytes_read;

  tidy_resource(config);

  return(length);
}


/*}}}  */

/*{{{  read resource file*/
int read_resource_file (config, systemname)
unsigned char config[];
char *systemname;

/*
 * Read resource configuration data from named file for named TCPlink system
 */
{
  int index = 0;
  unsigned char result;

  /*
   * Search by system name only
   * Requires name of local host, but as this is not required for this
   * search I'll just give it "localhost". The systemanme should not
   * be local!
   */

  result = CM_InitSearch("\0", systemname, "localhost");

  if (result == STATUS_NOERROR)
  {
    BOOL searching = TRUE;
    char tmpstring[MAX_NET_LINE_SIZE + 1]; /* plus one for '\0' */
    int len;
    int entries = 0;
    unsigned int link_no;

    init_data_block(config, RESOURCE_DATA_BLOCK);
  
    while (searching)
    {
      result = CM_DoSearch();

      if ((result == STATUS_NOERROR) && (entries == MAX_RESOURCE_ENTRIES))
      {
        searching = FALSE;
        printf("Error: too many resource entries for %s in %s\n", systemname,
                                                       getenv(RESOURCE_ENVIRONMENT));
      }
      else if (result == STATUS_NOERROR)
      {

        CM_GetResource(tmpstring);
        for (len = 0; tmpstring[len] != '\0'; len++)
          ; /* calculate length */
        config[index + RESOURCE_ALLOCATED_O] = (char) TRANSPUTER_FALSE;
        config[index + RESOURCE_LEN_O] = (char) len;
        strncpy(&config[index + RESOURCE_NAME_O], tmpstring, len);

        CM_GetLinkName (tmpstring);
#ifdef VMS  /* VMS sscanf bug */
        len = sscanf(tmpstring, "%d", &link_no);
#else
        len = sscanf(tmpstring, "%u", &link_no);
#endif
        if ((len == 1) && ((link_no >= 0) && (link_no < MAX_SUBSYSTEMS)))
        {
          config[index] = (unsigned char) (link_no & 0XFF);
          index = index + RESOURCE_ENTRY_SIZE;
          entries++;
        }
        else if (len == 1)
        {
          config[index] = (unsigned char) 0;
          searching = FALSE;
          printf("Error: link must be in range 0 - 3 in record ending at line %d of %s\n",
                 CM_GetLineNumber(), getenv(RESOURCE_ENVIRONMENT));
        
        }
        else
        {
          searching = FALSE;
          printf("Error: link not a numeric string in record ending at line %d of %s\n",
                 CM_GetLineNumber(), getenv(RESOURCE_ENVIRONMENT));
        }

      }
      else
      {
        if (result == STATUS_BAD_OPERATION)
        {
          printf("Error: %s\n", conlib_errorstring);
        }

        searching = FALSE;
      }
    }

    if (index == 0)
    {
      printf("Warning: no entries for %s found in %s\n", systemname,
                                                getenv(RESOURCE_ENVIRONMENT));
    }

    CM_EndSearch();
  }
  else
  {
    printf("Error: %s\n", conlib_errorstring);
  }
  return(index);
}
/*}}}  */

